import unittest

import os
import sys

sys.path += [os.path.abspath(__file__ + "/../../..")]

import os.path
from io import StringIO

from asp2cnl.compiler import compile_rule
from asp2cnl.parser import ASPParser

from cnl2asp.cnl2asp import Cnl2asp


class TestRules(unittest.TestCase):
    def test_simple(self):
        results = StringIO()

        program = open(os.path.join(os.path.dirname(__file__), "encoding_singleTest.asp"), "r").read()
        definitions = ASPParser(program).parse()

        #expectedCnl = open(os.path.join(os.path.dirname(__file__), "expected_output.cnl"), 'r').readlines()
        counterLine = 0

        with open(os.path.join(os.path.dirname(__file__), "schema_singleTest.cnl"), "r") as f:
            symbols = Cnl2asp(f).get_symbols()
            for rule in definitions:
                results.write("RULE: ")
                results.write("\n")
                results.write(rule.toString())
                print(rule.toString())
                results.write("\n")
                results.write("\n")
                results.write("TRANSLATED IN: ")
                results.write("\n")
                compiled = compile_rule(rule, symbols)
                print(compiled)
                #compiled = ""
                results.write(compiled)
                results.write("\n")
                #print(rule.toString())
                #print(compiled)
                outFileDisk = os.path.join(os.path.dirname(__file__), "output.cnl")
                with open(outFileDisk, "w") as out_file:
                    f.seek(0)
                    out_file.write(f.read())
                with open(outFileDisk, "a") as out_file:
                    out_file.write(compiled)
                    #self.assertEqual(compiled.rstrip(), expectedCnl[counterLine].rstrip(),
                    #                 "Expected CNL sentence '" + expectedCnl[counterLine] + "'; got '" + compiled + "'")
                # print("Translating: " + compiled)

                with open(outFileDisk, "r") as in_file:
                    cnl2asp = Cnl2asp(in_file)
                    result = cnl2asp.compile(False)
                    results.write("TRANSLATION BACK: ")
                    results.write("\n")
                    results.write(result)

                results.write("\n")
                results.write("------------------")
                counterLine = counterLine + 1
                results.write("\n")
            print("Results: \n")
            print(results.getvalue())


if __name__ == '__main__':
    unittest.main()



import argparse
import json

from asp2cnl.compiler import *
from asp2cnl.parser import ASPParser

from cnl2asp.cnl2asp import Cnl2asp


def run_asp2cnl():
    parser = argparse.ArgumentParser()
    parser.add_argument('--asp2cnl', action='store_true')
    parser.add_argument('input_file')
    parser.add_argument('definition_file')
    parser.add_argument('output_file', type=str, nargs='?', default='')
    parser.add_argument('--format_output_file', type=str, nargs='?', default='json')
    parser.add_argument("-s", '--silent', action='store_true')

    args = parser.parse_args()
    input_file = args.input_file
    output_file = args.output_file
    definition_file = args.definition_file
    silent = args.silent
    format_output_file = args.format_output_file

    program = open(input_file, "r").read()
    definitions = ASPParser(program).parse()

    with open(definition_file, "r") as f:
        symbols = Cnl2asp(f).get_symbols()
        o = {'asp': [], 'cnl': []}
        for rule in definitions:
            o['asp'].append(rule.toString().strip())
            compiled = compile_rule(rule, symbols)
            o['cnl'].append(compiled)
            if not silent:
                print(compiled)
    if output_file:
        try:
            with open(output_file, "w") as f:
                if format_output_file == "json":
                    json.dump(o, f)
                elif format_output_file == "standard":
                    for cnl in o['cnl']:
                        f.write(cnl + "\n")
                elif format_output_file == "aspcnl":
                    counter = 0
                    for cnl in o['cnl']:
                        f.write(o['asp'][counter] + "\n")
                        f.write("% " + cnl + "\n\n")
                        counter += 1


        except Exception as e:
            print("Error in writing output", str(e))



import unittest

from chemiotherapy_scheduling.test import TestRules as Test1
from graph_coloring.test import TestRules as Test2
#from hamiltonian_path.test import TestRules as Test3
from hanoi_tower.test import TestRules as Test4
#from integer_sets.test import TestRules as Test5
from mao.test import TestRules as TestMao
from maxclique.test import TestRules as Test6
from nurse_scheduling.test import TestRules as Test7
from robot_arm.test import TestRules as Test8
#from undirected_graph.test import TestRules as Test9

class TestEncodings(unittest.TestCase):
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test1))
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test2))
    #test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test3))
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test4))
    #test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test5))
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestMao))
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test6))
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test7))
    test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test8))
    #test_suite.addTest(unittest.TestLoader().loadTestsFromTestCase(Test9))


    runner=unittest.TextTestRunner()
    runner.run(test_suite)




import argparse
import json

from asp2cnl.compiler import compile_rule
from asp2cnl.parser import ASPParser
from cnl2nl.cnl2nl import compile_prompt, contact_llm_service

from cnl2asp.cnl2asp import Cnl2asp


def run_asp2nl():
    parser = argparse.ArgumentParser()
    parser.add_argument('--asp2nl', action='store_true')
    parser.add_argument("-m", "--llm_model", type=str, default='openchat')
    parser.add_argument('input_file')
    parser.add_argument('definition_file')
    parser.add_argument('output_file', type=str, nargs='?', default='')
    args = parser.parse_args()

    input_file = args.input_file
    definition_file = args.definition_file
    llm_model = args.llm_model
    output_file = args.output_file

    program = open(input_file, "r").read()
    definitions = ASPParser(program).parse()
    prompt_compiler = compile_prompt()

    o = {'asp': [], 'cnl': []}
    with open(definition_file, "r") as f:
        symbols = Cnl2asp(f).get_symbols()
        for rule in definitions:
            o['asp'].append(rule.toString())
            o['cnl'].append(compile_rule(rule, symbols))

    o['nl'] = []
    for cnl in o['cnl']:
        print(f'Processing rule: {cnl}')
        o['nl'].append(contact_llm_service(model=llm_model, prompt=prompt_compiler(cnl)))

    print("\n** NL:")
    print('\n'.join(o['nl']))
    try:
        with open(output_file, 'w') as f:
            json.dump(o, f)
    except Exception as e:
        print("Error in writing output", str(e))



from dataclasses import dataclass
from io import StringIO

import sys
import os
##ROOT_CNL2ASP_PATH = 'C:/Users/Kristian/git/cnl/cnl2asp/'
##sys.path += [ROOT_CNL2ASP_PATH + 'src']

from cnl2asp.cnl2asp import Symbol, SymbolType

from asp2cnl.parser import Directive, ClassicalLiteral, BuiltinAtom, NafLiteral, AggregateLiteral, Term, ArithmeticAtom


class TermDecorator(Term):
    sentenceBefore:str = None
    sentenceAfter: str = None
    literal = None
    literalVariable: str = None
    def toString(self):
        sentence = super().toString()
        if self.sentenceBefore is not None:
            sentence = self.sentenceBefore + " " + sentence
        if self.sentenceAfter is not None:
            sentence = sentence + " " + self.sentenceAfter
        if self.literalVariable is not None:
            sentence = sentence + " " + self.literalVariable

        return sentence


def extract_name(name):
    if type(name) == Symbol:
        return extract_name(name.predicate)
    return name


def get_symbol(symbols, atom):
    symbol_name = atom.name
    # print(symbol_name)
    # symbol_name = symbol_name.replace("_", " ")
    res: list = [symbols[i] for i in
                 range(len(symbols)) if
                 symbols[i].predicate.lower() == symbol_name.lower()]

    if len(res) == 0:
        return None
    else:
        symb = None

        for s in res:
            # print(s)
            if s.symbol_type == SymbolType.TEMPORAL:
                s.attributes = s.attributes[0:1]
            # print(len(s.attributes))
            # print(atom.arity())
            if len(s.attributes) == atom.arity():
                symb = s

        for i in range(len(symb.attributes)):
            if type(symb.attributes[i]) == Symbol:
                if symb.attributes[i].symbol_type == SymbolType.DEFAULT:
                    symb.attributes[i] = symb.attributes[i].predicate.strip() + " " + (
                        symb.attributes[i].attributes[0].strip()).lower()
                else:
                    symb.attributes[i] = symb.attributes[i].predicate.strip()
                ##print("Cosa")
                ##print(res[0].attributes[i])
                ##res[0].attributes[i] = str(res[0].attributes[i].attributes[0])
            else:
                symb.attributes[i] = symb.attributes[i].strip()
        # print (symb)
        return symb


def compile_rule(rule, symbols):
    results = StringIO()
    if type(rule) == Directive:
        results.write(generate_directive(rule, symbols))
        #results.write("\n")
    else:
        if rule.isFact():
            # Facts
            atom = rule.head.atoms[0]
            symb = get_symbol(symbols, atom)
            if symb is None or symb.symbol_type == SymbolType.DEFAULT:
                if len(atom.terms) == 1:
                    if symb is None:
                        results.write(generate_is_a(atom))
                    else:
                        if atom.terms[0].isWithDotDot():
                            results.write(generate_goes(atom))
                        else:
                            results.write(generate_there_is(None, atom, symb, {}, True))
                    results.write(".")
                    #results.write("\n")

                elif len(atom.terms) >= 2:
                    if symb is not None:
                        results.write(generate_there_is(None, atom, symb, {}, True))
                        results.write(".")
                        #results.write("\n")
                        # else:
                    #    results.write(generate_relation(atom))
                    #    results.write("\n")
        elif rule.isClassical():
            results.write(generate_classical_statement(rule, symbols))
            #results.write("\n")
        elif rule.isStrongConstraint():
            results.write(generate_strong_constraint(rule, symbols))
            #results.write("\n")
        elif rule.isDisjunctive() or rule.isChoice():
            results.write(generate_disjunctive_or_choice_statement(rule, symbols))
            #results.write("\n")
        elif rule.isWeakConstraint():
            results.write(generate_weak_constraint(rule, symbols))
            #results.write("\n")

    return results.getvalue()


def generate_directive(directive, symbols):
    results = StringIO()
    if directive.type == "const":
        results.write(directive.name)
        results.write(" is a constant equal to")
        results.write(" ")
        results.write(directive.value)
        results.write(".")
    return results.getvalue()


def generate_is_a(atom):
    # Eg. pub(1). --> 1 is a pub.
    results = StringIO()
    results.write(atom.terms[0].name.replace('"', '').capitalize())
    results.write(" ")
    results.write("is a")
    results.write(" ")
    results.write(atom.name)
    return results.getvalue()


def generate_goes(atom):
    results = StringIO()
    results.write("A")
    results.write(" ")
    results.write(atom.name)
    results.write(" ")
    results.write("goes")
    results.write(" ")
    results.write("from")
    results.write(" ")
    results.write(atom.terms[0].name)
    results.write(" ")
    results.write("to")
    results.write(" ")
    results.write(atom.terms[0].afterDotDot)
    return results.getvalue()


def generate_there_is(body, atom, symbol, builtinAtoms, start=False, noThereIs=False, literalVariables=None):
    # Eg. movie(1,"jurassicPark","spielberg",1993).
    # -->
    # There is a movie with id equal to 1, with director equal to spielberg, with title equal to jurassicPark, with year equal to 1993.
    isNot = False
    if type(atom) == NafLiteral:
        isNot = atom.isNot
        atom = atom.literal
    results = StringIO()
    if not noThereIs:
        if start:
            results.write("There")
        else:
            results.write("there")
        results.write(" ")
        results.write("is")
        results.write(" ")
    if isNot:
        results.write("not")
        results.write(" ")
    results.write("a")
    results.write(" ")
    results.write(atom.name)
    results.write(" ")
    doneOfTheIdentifierTerms = []
    if body is not None:
        for builtin in builtinAtoms:
            #if builtin not in doneOfTheIdentifier:
                for term in builtin.terms:
                    if isinstance(term, TermDecorator):
                        if term.literal.literal == atom and term.literalVariable not in doneOfTheIdentifierTerms:
                            #results.write(get_literal_identifier(body, term.literal))
                            results.write(term.literalVariable)
                            results.write(" ")
                            #doneOfTheIdentifier = True
                            doneOfTheIdentifierTerms.append(term.literalVariable)

    #if literalVariables is not None and not doneOfTheIdentifier:
    if literalVariables is not None:
        for literalVariable in literalVariables:
            if literalVariable not in doneOfTheIdentifierTerms:
                results.write(literalVariable)
                results.write(" ")

    results.write(generate_with(atom, symbol, builtinAtoms))
    return results.getvalue()


def get_literal_identifier(body, lit):
    count = 0
    for l in body.literals:
        if type(l) == NafLiteral and type(l.literal) == ClassicalLiteral:
            if l.literal.name == lit.literal.name:
                count = count + 1
                if l == lit:
                    return lit.literal.name.upper() + str(count)
    return None


def generate_vars_symbols(body, symbols, arithAtom):
    results = StringIO()
    matchedVars = []
    for term in arithAtom.terms:
        if term.isVariable():
            matchedVars.append((term, None, None))
        else:
            matchedVars.append(term)

    foundLitNames = []
    for lit in body.literals:
        if type(lit) == NafLiteral and type(lit.literal) == ClassicalLiteral:
            symbLit = get_symbol(symbols, lit.literal)
            atom = lit.literal
            foundLitNames.append(atom.name)
            for i in range(len(atom.terms)):
                canContinue = True
                if type(atom.terms[i]) == Term:
                    if atom.terms[i].isUnderscore():
                        canContinue = False
                if canContinue:
                    for ip in range(len(matchedVars)):
                        if type(matchedVars[ip]) == tuple:
                            if matchedVars[ip][0] == atom.terms[i]:
                                p = list(matchedVars[ip])
                                p[1] = symbLit.attributes[i]
                                p[2] = lit
                                #                                foundLitNames.append(lit.literal.name)
                                matchedVars[ip] = tuple(p)
        elif type(lit) == AggregateLiteral:
            operator = getAggregateOperator(lit)
            assignmentVar = None
            if operator == "=":
                if (lit.lowerOp == "=" and lit.upperGuard is None) or (
                        lit.upperOp == "=" and lit.lowerGuard is None):
                    if lit.upperGuard is not None and lit.upperGuard.isVariable():
                        assignmentVar = lit.upperGuard
                    elif lit.lowerGuard is not None and lit.lowerGuard.isVariable():
                        assignmentVar = lit.lowerGuard
            if assignmentVar is not None:
                for ip in range(len(matchedVars)):
                    if type(matchedVars[ip]) == tuple:
                        if matchedVars[ip][0] == assignmentVar:
                            p = list(matchedVars[ip])
                            #p[1] = symbLit.attributes[i]
                            p[2] = lit
                            #                                foundLitNames.append(lit.literal.name)
                            matchedVars[ip] = tuple(p)



    started = False
    for builtVars in matchedVars:
        isConstant = type(builtVars) != tuple
        if started:
            results.write(", ")
            if isConstant or builtVars == matchedVars[-1]:
                results.write("and ")
        else:
            started = True
        if not isConstant:
            #print(builtVars)
            if type(builtVars[2]) == AggregateLiteral:
                results.write(builtVars[0].name)
            else:
                if builtVars[2] is not None:
                    if foundLitNames.count(builtVars[2].literal.name) == 1:
                        results.write(builtVars[0].name)
                    else:
                        results.write("the")
                        results.write(" ")
                        results.write(builtVars[1])

                        results.write(" ")
                        results.write(builtVars[0].name)

                        results.write(" ")
                        results.write("of the ")
                        results.write(builtVars[2].literal.name)
                        results.write(" ")
                        results.write(get_literal_identifier(body, builtVars[2]))
                        # results.write(builtVars[2].upper())
                else:
                    results.write(builtVars[0].name)
        else:
            results.write(builtVars.toString())

    return results.getvalue()


def generate_with(atom, symbol, builtinAtoms=[]):
    results = StringIO()
    started = False
    for i in range(len(atom.terms)):
        canContinue = True
        if type(atom.terms[i]) == Term:
            if atom.terms[i].isUnderscore():
                canContinue = False
        if canContinue:
            if started:
                results.write(", ")
            else:
                started = True
                # results.write(" ")
            # if symbol.symbol_type == SymbolType.DEFAULT:
            results.write("with")
            results.write(" ")
            results.write(symbol.attributes[i])
            results.write(" ")
            if type(atom.terms[i]) == Term:
                if atom.terms[i].isVariable():
                    foundMatchedBuiltin = False
                    userVariablesInBuiltin = []
                    for builtinAtom in builtinAtoms:
                        # if atom.terms[i] in builtinAtoms.keys():
                        if not type(builtinAtom.terms[0]) == ArithmeticAtom:
                            if (builtinAtom.terms[0] == atom.terms[i]
                                    and builtinAtom.terms[0] not in userVariablesInBuiltin):
                                userVariablesInBuiltin.append(builtinAtom.terms[0])
                                foundMatchedBuiltin = True
                                results.write(atom.terms[i].name)
                                results.write(" ")
                                # builtinAtom = builtinAtoms[atom.terms[i]]
                                results.write(generate_compare_operator_sentence(builtinAtom.op))
                                results.write(" ")

                                results.write(builtinAtom.terms[1].toString())

                                builtinAtoms.remove(builtinAtom)

                    if not foundMatchedBuiltin:
                        results.write(atom.terms[i].name)

                else:
                    results.write("equal")
                    results.write(" ")
                    results.write("to")
                    results.write(" ")
                    results.write(atom.terms[i].name.strip('\"'))
            elif type(atom.terms[i]) == ArithmeticAtom:
                '''results.write("equal")   
                results.write(" ")
                results.write("to")   
                results.write(" ")'''
                results.write(atom.terms[i].toString())
                # results.write(extract_name(symbol.attributes[i]))
        # results.write(" ")
        # results.write("equal to")
        # results.write(" ")
        # results.write(atom.terms[i].name.strip('\"'))
    return results.getvalue()


def generate_compare_operator_sentence(operator):
    results = StringIO()
    if operator == "!=" or operator == "<>":
        # different from
        results.write("different")
        results.write(" ")
        results.write("from")
    elif operator == "<":
        # less than
        results.write("less")
        results.write(" ")
        results.write("than")
    elif operator == "<=":
        # less than or equal to
        results.write("less")
        results.write(" ")
        results.write("than")
        results.write(" ")
        results.write("or")
        results.write(" ")
        results.write("equal")
        results.write(" ")
        results.write("to")
    elif operator == "=":
        # equal to
        results.write("equal")
        results.write(" ")
        results.write("to")
    elif operator == ">":
        # greater than
        results.write("greater")
        results.write(" ")
        results.write("than")
    elif operator == ">=":
        # greater than or equal to
        results.write("greater")
        results.write(" ")
        results.write("than")
        results.write(" ")
        results.write("or")
        results.write(" ")
        results.write("equal")
        results.write(" ")
        results.write("to")
    elif operator == "between":
        results.write("between")
    return results.getvalue()


# TODO

def generate_relation(atom):
    # Eg. work_in("john",1).
    # --> Waiter John works in pub 1.
    # serve("john","alcoholic").
    # --> Waiter John serves a drink alcoholic.
    results = StringIO()
    # symb = get_symbol(symbols, atom.name.replace("_", " "))
    # atom_name = atom.name.replace("_", " ")
    atom_name = atom.name

    # symb = get_symbol(symbols, atom.terms[0])
    # if symb is not None:
    # results.write(symb.predicate('"', '').capitalize())
    # results.write(" ")
    results.write(atom.terms[0].name.replace('"', ''))
    results.write(" ")
    results.write(atom_name)
    for i in range(len(atom.terms)):
        if i > 0:
            results.write(" ")
            results.write("and")
        results.write(" ")
        results.write(atom.terms[i].name.replace('"', ''))
        # symb2 = get_symbol(symbols, atom.terms[1])

    '''
    for i in range(len(atom.terms)):
        symb = get_symbol(symbols, atom.name)
        if symb is not None:            
            results.write(atom.terms[i].replace('"', '').capitalize())

    results.write(symb.predicate)
    results.write(" ")
    results.write(atom.terms[1].replace('"', ''))
    '''
    results.write(".")

    return results.getvalue()


def generate_disjunctive_or_choice_statement(rule, symbols):
    # DISJUNCTIVE
    # Eg. scoreassignment(movie(I),1) | scoreassignment(movie(I),2)
    #               | scoreassignment(movie(I),3) :- movie(I,_,_,_).
    # -->
    # Whenever there is a movie with id I, we can have with director equal to spielberg
    # a scoreAssignment with movie I, and with value equal to 1
    # or a scoreAssignment with movie I, and with value equal to 2
    # or a scoreAssignment with movie I, and with value equal to 3.
    # -------
    # CHOICE
    # Eg. 0 <= {topmovie(I):movie(I,_,X,_)} <= 1 :- director(X), X != spielberg.
    # -->
    # Whenever there is a director with name X different from spielberg
    #         then we can have at most 1 topmovie with id I such that there is a movie with director X,
    #           and with id I.
    results = StringIO()

    builtinAtoms = getBuiltinAtoms(rule, symbols)
    results.write(generate_body(rule, symbols, builtinAtoms))
    results.write(" ")
    results.write("then")
    results.write(" ")
    results.write("we")
    results.write(" ")
    results.write("can")
    results.write(" ")
    results.write("have")
    results.write(" ")

    if rule.isChoice():
        results.write(generate_head_choice(rule.head, symbols))
    else:
        results.write(generate_head(rule.head, symbols))

    results.write(generateWhereForBuiltins(builtinAtoms))
    results.write(".")
    return results.getvalue()


def getBuiltinAtoms(rule, symbols):
    builtinAtoms = []
    if rule.isChoice():
        for choice_element in rule.head.elements:
            if choice_element.right_part is not None:
                for nafLit in choice_element.right_part:
                    if isinstance(nafLit.literal, BuiltinAtom):
                        builtinAtoms.append(nafLit.literal)

    for lit in rule.body.literals:
        if type(lit) == NafLiteral and type(lit.literal) == BuiltinAtom:
            builtinAtoms.append(lit.literal)

    # Angle Management
    foundAngleAtoms = set()
    for builtinAtom in builtinAtoms:
        checkNumber = 1
        while checkNumber < 3:
            leftPart = None
            rithPart = None
            if checkNumber == 1:
                leftPart = builtinAtom.terms[0]
                rithPart = builtinAtom.terms[1]
            elif checkNumber == 2:
                leftPart = builtinAtom.terms[1]
                rithPart = builtinAtom.terms[0]
            if isinstance(rithPart, ArithmeticAtom):
                arithAtom = rithPart
                removedFromVariable = False
                if isinstance(leftPart, ArithmeticAtom):
                    if len(leftPart.ops) == 1:
                        if isinstance(leftPart.terms[1], Term):
                            if ( (leftPart.ops[0] == "/" or leftPart.ops[0] == "\\")
                                            and leftPart.terms[1].name == "360" ) :
                                leftPart = leftPart.terms[0]
                                removedFromVariable = True

                if arithAtom.terms[-1].name == "360" and leftPart.isVariable():
                    checkingVar = leftPart
                    for lit in rule.body.literals:
                        if type(lit) == NafLiteral and type(lit.literal) == ClassicalLiteral:
                            symbLit = get_symbol(symbols, lit.literal)
                            atom = lit.literal
                            isAngle = False
                            angleDescription = None
                            if atom.name == "angle":
                                isAngle = True
                                angleDescription = "angle"
                            else:
                                for i in range(len(atom.terms)):
                                    if checkingVar == atom.terms[i]:
                                        if symbLit.attributes[i] == "angle value" \
                                                or symbLit.attributes[i] == "angle_value" \
                                                or symbLit.attributes[i].endswith(" angle_value") \
                                                or symbLit.attributes[i].endswith(" angle value") \
                                                or symbLit.attributes[i].endswith("_angle") \
                                                or symbLit.attributes[i].endswith(" angle"):
                                            isAngle = True
                                            foundAngleAtoms.add(lit)
                                            angleDescriptionBefore = "the " + symbLit.attributes[i]
                                            angleDescriptionAfter = "of the " + symbLit.predicate
                                            angleLiteral = lit
                                            angleLiteralVariable = atom.name.upper() + "_WITH_ANGLE_A" + str(len(foundAngleAtoms))

                            if isAngle:
                                if len(arithAtom.terms) == 2:
                                    if checkNumber == 1:
                                        #builtinAtom.terms[1] = TermDecorator(arithAtom.terms[0].name, arithAtom.terms[0].afterDotDot, arithAtom.terms[0].isModule, angleDescription)
                                        builtinAtom.terms[1] = arithAtom.terms[0]
                                        if removedFromVariable:
                                            #builtinAtom.terms[0] = leftPart
                                            builtinAtom.terms[0] = TermDecorator(leftPart.name, leftPart.afterDotDot, leftPart.isModule)
                                            builtinAtom.terms[0].sentenceBefore = angleDescriptionBefore
                                            builtinAtom.terms[0].sentenceAfter = angleDescriptionAfter
                                            builtinAtom.terms[0].literal = angleLiteral
                                            builtinAtom.terms[0].literalVariable = angleLiteralVariable
                                    else:
                                        builtinAtom.terms[0] = arithAtom.terms[0]
                                        if removedFromVariable:
                                            #builtinAtom.terms[1] = leftPart
                                            builtinAtom.terms[1] = TermDecorator(leftPart.name, leftPart.afterDotDot, leftPart.isModule)
                                            builtinAtom.terms[1].sentenceBefore = angleDescriptionBefore
                                            builtinAtom.terms[1].sentenceAfter = angleDescriptionAfter
                                            builtinAtom.terms[1].literal = angleLiteral
                                            builtinAtom.terms[1].literalVariable = angleLiteralVariable
                                else:
                                    arithAtom.terms.pop()
                                    arithAtom.ops.pop()
                                checkNumber = 3
            checkNumber = checkNumber + 1

    return builtinAtoms


def generate_classical_statement(rule, symbols):
    # Eg. topmovie(X) :- movie(X,_,"spielberg",_).
    # -->
    # Whenever there is a movie with id X, with director equal to spielberg
    # then we must have a topmovie with id X.
    results = StringIO()
    builtinAtoms = getBuiltinAtoms(rule, symbols)
    results.write(generate_body(rule, symbols, builtinAtoms))
    results.write(" ")
    results.write("then")
    results.write(" ")
    results.write("we")
    results.write(" ")
    results.write("must")
    results.write(" ")
    results.write("have")
    results.write(" ")

    results.write(generate_head(rule.head, symbols))
    results.write(generateWhereForBuiltins(builtinAtoms))
    results.write(".")
    return results.getvalue()


def generate_strong_constraint(rule, symbols):
    # Eg. :- movie(X,_,_,1964), topmovie(Y), X = Y.
    # -->
    # It is prohibited that X is equal to Y, whenever there is a movie
    # with id X, and with year equal to 1964, whenever there is a topMovie with id Y.
    results = StringIO()
    results.write("It is prohibited that")
    builtinAtoms = getBuiltinAtoms(rule, symbols)
    results.write(generate_body(rule, symbols, builtinAtoms, True))
    results.write(generateWhereForBuiltins(builtinAtoms))
    results.write(".")
    return results.getvalue()


def generate_weak_constraint(rule, symbols):
    results = StringIO()
    results.write("It is preferred")
    if type(rule.weight_at_level.beforeAt) != ArithmeticAtom:
        if rule.weight_at_level.beforeAt.name == "1":
            results.write(" as little as possible")

    if rule.weight_at_level.afterAt is not None:
        results.write(", ")
        results.write("with")
        results.write(" ")
        # if rule.weight_at_level.afterAt[0].name == "1":
        #    results.write("low")
        #    results.write(" ")
        # elif rule.weight_at_level.afterAt[0].name == "2":
        #    results.write("medium")
        #    results.write(" ")
        # elif rule.weight_at_level.afterAt[0].name == "3":
        #    results.write("high")
        #    results.write(" ")
        results.write("priority")
        results.write(" ")
        results.write(rule.weight_at_level.afterAt[0].name)
        results.write(",")
    results.write(" ")
    results.write("that")
    # results.write(" ")

    builtinAtoms = getBuiltinAtoms(rule, symbols)
    results.write(generate_body(rule, symbols, builtinAtoms, False, rule.weight_at_level.beforeAt))

    # results.write(", ")
    # results.write(rule.weight_at_level.beforeAt.name)
    # results.write(" ")
    '''if type(rule.weight_at_level.beforeAt) != ArithmeticAtom and len(foundAggrs) > 0:
        if rule.weight_at_level.beforeAt.isVariable():
            results.write("is")
            results.write(" ")    
            if rule.weight_at_level.isMaximize:        
                results.write("maximized")
            else:
                results.write("minimized")'''
    results.write(generateWhereForBuiltins(builtinAtoms))
    results.write(".")
    return results.getvalue()


def generate_head_choice(head, symbols):
    results = StringIO()
    atMost = False
    atLeast = False
    beetween = False
    exactly = False

    howMany = StringIO()

    if head.upperGuard is not None:
        if head.upperOp == "<=":
            if head.lowerGuard is None:
                atMost = True
            else:
                if head.lowerOp == "<=":
                    if head.lowerGuard.name == head.upperGuard.name:
                        exactly = True
                    elif int(head.lowerGuard.name) < int(head.upperGuard.name):
                        beetween = True
        elif head.upperOp == "=":
            if head.lowerGuard is None:
                exactly = True
            else:
                if head.lowerOp == "=":
                    if head.lowerGuard.name == head.upperGuard.name:
                        exactly = True
    else:
        if head.lowerGuard is not None:
            if head.lowerOp == "=":
                exactly = True
            elif head.lowerOp == "<=":
                atLeast = True
    if exactly:
        # exactly 1 topmovie with id I such that there is a movie with director X,
        #           and with id I.
        howMany.write("exactly")
        howMany.write(" ")
        if head.lowerGuard is not None:
            howMany.write(head.lowerGuard.name)
        else:
            howMany.write(head.upperGuard.name)
    if atLeast:
        # at least 1 topmovie with id I such that there is a movie with director X,
        #           and with id I.
        howMany.write("at least")
        howMany.write(" ")
        howMany.write(head.lowerGuard.name)
    if atMost:
        # at most 1 topmovie with id I such that there is a movie with director X,
        #           and with id I.
        howMany.write("at most")
        howMany.write(" ")
        howMany.write(head.upperGuard.name)
    if beetween:
        # between 3 and 4 topmovie with id I such that there is a movie with director X,
        #           and with id I.
        howMany.write("between")
        howMany.write(" ")
        howMany.write(head.lowerGuard.name)
        howMany.write(" ")
        howMany.write("and")
        howMany.write(" ")
        howMany.write(head.upperGuard.name)
    if head.lowerGuard is None and head.upperGuard is None:
        howMany.write("a")

    startedHeadElem = False
    for headElem in head.elements:
        if startedHeadElem:
            results.write(" ")
            results.write("or")
            # results.write(" ")
        else:
            startedHeadElem = True
            results.write(howMany.getvalue())
        results.write(" ")
        results.write(headElem.left_part.name)
        results.write(" ")
        results.write(generateWithInHead(symbols, headElem.left_part))
        if headElem.right_part is not None:
            results.write(" ")
            results.write("such that")
            results.write(" ")

            started = False
            for nafLit in headElem.right_part:
                if type(nafLit.literal) != BuiltinAtom:
                    if started:
                        results.write(",")
                        results.write(" ")
                    symb = get_symbol(symbols, nafLit.literal)
                    results.write(generate_there_is(None, nafLit, symb, {}, False, started))
                    started = True
    #results.write(", where J1 is less than J2 ")

    return results.getvalue()


def generate_head(head, symbols):
    results = StringIO()
    started = False
    for atom in head.atoms:
        if started:
            results.write(" ")
            results.write("or")
            results.write(" ")
        else:
            started = True
        results.write("a")
        results.write(" ")
        results.write(atom.name)
        results.write(" ")

        results.write(generateWithInHead(symbols, atom))

    return results.getvalue()


def generateWhereForBuiltins(builtinAtoms):
    results = StringIO()
    if len(builtinAtoms) > 0:
        # started = False
        for builtinAtom in builtinAtoms:
            #if type(builtinAtom.terms[0]) != ArithmeticAtom:
            # if started:
            results.write(", ")
            # else:
            #    started = True
            results.write("where")
            results.write(" ")
            results.write(builtinAtom.terms[0].toString())
            results.write(" ")
            results.write("is")
            results.write(" ")
            results.write(generate_compare_operator_sentence(builtinAtom.op))
            results.write(" ")
            results.write(builtinAtom.terms[1].toString())
    return results.getvalue()


def generateWithInHead(symbols, atom):
    results = StringIO()
    symbLit = get_symbol(symbols, atom)

    for i in range(len(symbLit.attributes)):
        if i > 0:
            results.write(",")
            results.write(" ")
        results.write("with")
        results.write(" ")
        results.write(symbLit.attributes[i])
        results.write(" ")

        if type(atom.terms[i]) == Term and not atom.terms[i].isVariable():
            results.write("equal")
            results.write(" ")
            results.write("to")
            results.write(" ")
        results.write(atom.terms[i].toString().strip('\"'))
        # results.write(" ")
    return results.getvalue()


def generate_body(rule, symbols, builtinAtoms, isStrongConstraint=False, costWeakTerm=None):
    body = rule.body
    results = StringIO()
    startedLits = False
    # builtinAtoms = []
    hasSumInBuiltin = False
    builtinSumSetInConstraint = False

    literalVariables = []
    for builtinA in builtinAtoms:
        if type(builtinA.terms[0]) == ArithmeticAtom:
            hasSumInBuiltin = True
            for term in builtinA.terms:
                if isinstance(term, TermDecorator):
                    for lit in body.literals:
                        if term.literal == lit:
                            literalVariables.append(term.literalVariable)

    if hasSumInBuiltin and isStrongConstraint:
        # results.write(" ")
        # started = False
        for builtinAtom in builtinAtoms:
            if type(builtinAtom.terms[0]) == ArithmeticAtom:
                if startedLits:
                    results.write(", whenever we have that")
                else:
                    builtinSumSetInConstraint = True
                    startedLits = True
                results.write(" ")
                results.write(generate_operation_between(body, symbols, builtinAtom.terms[0]))
                results.write(generate_compare_of_arithmetic_builtin(builtinAtom.op, builtinAtom.terms[1]))
                builtinAtoms.remove(builtinAtom)

    tmpWheneverResults = None
    tmpAggrResults = None
    foundAggrs = []
    firstConstraintLiteralInSentence = None
    specialConstraintTranslationForLiteral = False

    # It
    foundLitNames = []
    for lit in body.literals:
        if type(lit) == NafLiteral and type(lit.literal) == ClassicalLiteral:
            foundLitNames.append(lit.literal.name)

    for lit in body.literals:
        if type(lit) == NafLiteral and type(lit.literal) == ClassicalLiteral:
            if tmpWheneverResults is None:
                tmpWheneverResults = StringIO()
            if startedLits:
                tmpWheneverResults.write(",")
                tmpWheneverResults.write(" ")
                tmpWheneverResults.write("whenever")
            else:
                if not isStrongConstraint:
                    if len(foundAggrs) > 0:
                        tmpWheneverResults.write(", ")
                        tmpWheneverResults.write("whenever")
                        #results.write(", ")
                        #results.write("whenever")
                    elif costWeakTerm is not None:
                        if type(rule.weight_at_level.beforeAt) == ArithmeticAtom or rule.weight_at_level.beforeAt.isVariable():
                            tmpWheneverResults.write(" ")
                            tmpWheneverResults.write("whenever")
                    else:
                        tmpWheneverResults.write("Whenever")
                        #results.write("Whenever")
                else:
                    if len(foundAggrs) > 0 or hasSumInBuiltin:
                        tmpWheneverResults.write(" ")
                        tmpWheneverResults.write("whenever")
                    else:
                        specialConstraintTranslationForLiteral = True

                # if costWeakTerm is not None:
                #    tmpWheneverResults.write(" ")
                #    tmpWheneverResults.write("whenever")
                # elif not isStrongConstraint:
                #    tmpWheneverResults.write("Whenever")
                startedLits = True
            symbLit = get_symbol(symbols, lit.literal)
            if not specialConstraintTranslationForLiteral:
                tmpWheneverResults.write(" ")
                hasSum = hasSumInBuiltin
                #literalVariable = None
                if type(costWeakTerm) == ArithmeticAtom:
                    hasSum = True
                    if foundLitNames.count(lit.literal.name) > 1:
                        # Using literal indentifier for multiple atoms with same name
                        literalVariables.append(get_literal_identifier(body, lit))

                #for builtinAtom in builtinAtoms:
                #    for term in builtinAtom.terms:
                #        if isinstance(term, TermDecorator):
                #            if term.literal == lit:
                #                literalVariable = term.literalVariable

                tmpWheneverResults.write(generate_there_is(body, lit, symbLit, builtinAtoms, literalVariables=literalVariables))
            else:
                specialConstraintTranslationForLiteral = False
                if not hasSumInBuiltin:
                    firstConstraintLiteralInSentence = " " + generate_there_is(body, lit, symbLit, builtinAtoms)
            # tmpWheneverResults.write(" ")
        elif type(lit) == AggregateLiteral:
            if not isStrongConstraint and costWeakTerm is None:
                if startedLits and len(foundAggrs) == 0:
                    results.write(tmpWheneverResults.getvalue())
                    tmpWheneverResults = StringIO()
                    results.write(",")

                if len(foundAggrs) > 0 or startedLits:
                    results.write(" ")
                    results.write("whenever")
                else:
                    results.write("Whenever")
                results.write(" ")
                results.write("we have that")
            else:
                if len(foundAggrs) > 0 or builtinSumSetInConstraint:  # or startedLits:
                    results.write(",")
                    results.write(" ")
                    results.write("whenever")
                    results.write(" ")
                    results.write("we have that")
            results.write(" ")
            results.write(generate_aggregate_subsentence(lit, symbols, costWeakTerm, isStrongConstraint))
            foundAggrs.append(lit)

    needVariable = False

    if len(foundAggrs) == 0:
        needVariable = True
    else:
        for foundAggr in foundAggrs:
            if (foundAggr.lowerOp == "=" and foundAggr.upperGuard is None) or (
                    foundAggr.upperOp == "=" and foundAggr.lowerGuard is None):
                if foundAggr.upperGuard is not None and foundAggr.upperGuard.isVariable():
                    needVariable = True
                elif foundAggr.lowerGuard is not None and foundAggr.lowerGuard.isVariable():
                    needVariable = True

    if tmpWheneverResults is not None:
        if costWeakTerm is not None and needVariable:
            if type(costWeakTerm) == ArithmeticAtom or len(foundAggrs) > 0:
                if type(costWeakTerm) == ArithmeticAtom:
                    results.write(" ")
                    results.write(generate_operation_between(body, symbols, costWeakTerm))
                    # if rule.weight_at_level.beforeAt.isVariable():
                    results.write(" ")
                results.write("is")
                results.write(" ")
                if rule.weight_at_level.isMaximize:
                    results.write("maximized")
                else:
                    results.write("minimized")
            else:
                if rule.weight_at_level.beforeAt.isVariable():
                    tmpWheneverResults.write(", ")
                    tmpWheneverResults.write(costWeakTerm.name)
                    tmpWheneverResults.write(" ")

        if firstConstraintLiteralInSentence is not None:
            if len(foundAggrs) > 0:
                results.write(" whenever " + firstConstraintLiteralInSentence)
            else:
                results.write(firstConstraintLiteralInSentence)
        results.write(tmpWheneverResults.getvalue())

    if costWeakTerm is not None:  # and needVariable:
        if ((type(rule.weight_at_level.beforeAt) != ArithmeticAtom) and
                (len(foundAggrs) == 0 or (len(foundAggrs) > 0 and len(body.literals) == 1))):
            if rule.weight_at_level.beforeAt.isVariable():
                results.write("is")
                results.write(" ")
                if rule.weight_at_level.isMaximize:
                    results.write("maximized")
                else:
                    results.write("minimized")

    return results.getvalue()


def getAggregateOperator(aggregate):
    operator = None
    if (aggregate.lowerOp == "=" and aggregate.upperGuard is None
            or aggregate.upperOp == "=" and aggregate.lowerGuard is None
            or aggregate.lowerOp == "=" and aggregate.upperOp == "="
            and aggregate.lowerGuard.name == aggregate.upperGuard.name
            or aggregate.lowerOp == "<=" and aggregate.upperOp == "<="
            and aggregate.lowerGuard.name == aggregate.upperGuard.name
            or aggregate.lowerOp == ">=" and aggregate.upperOp == ">="
            and aggregate.lowerGuard.name == aggregate.upperGuard.name
    ):
        operator = "="
        if (aggregate.lowerOp == "=" and aggregate.upperGuard is None) or (
                aggregate.upperOp == "=" and aggregate.lowerGuard is None):
            if aggregate.upperGuard is not None and aggregate.upperGuard.isVariable():
                assignmentVar = aggregate.upperGuard
            elif aggregate.lowerGuard is not None and aggregate.lowerGuard.isVariable():
                assignmentVar = aggregate.lowerGuard

    elif (aggregate.upperOp == ">" and aggregate.lowerGuard is None):
        operator = ">"
    elif (aggregate.upperOp == ">=" and aggregate.lowerGuard is None):
        operator = ">="
    elif (aggregate.upperOp == "<" and aggregate.lowerGuard is None):
        operator = "<"
    elif (aggregate.upperOp == "<=" and aggregate.lowerGuard is None):
        operator = "<="
    elif (aggregate.lowerOp == "<=" and aggregate.upperOp == "<="):
        operator = "between"
    elif ((aggregate.lowerOp == "!=" or aggregate.lowerOp == "<>") and aggregate.upperGuard is None
          or (aggregate.upperOp == "!=" or aggregate.upperOp == "<>") and aggregate.lowerGuard is None
          or (aggregate.lowerOp == "!=" or aggregate.lowerOp == "<>") and (
                  aggregate.upperOp == "!=" or aggregate.upperOp == "<>")
    ):
        operator = "!="
        #    results.write("beet")
    return operator


def generate_operation_between(body, symbols, arithAtom):
    results = StringIO()
    isSum = False
    isSub = False
    isMult = False
    isDiv = False
    if arithAtom.ops[0] == "+":
        isSum = True
    if arithAtom.ops[0] == "-":
        isSub = True
    if arithAtom.ops[0] == "*":
        isMult = True
    if arithAtom.ops[0] == "/" or arithAtom.ops[0] == "\\":
        isDiv = True

    for op in arithAtom.ops:
        if op == "+":
            isSub = False
            isMult = False
            isDiv = False
        if op == "-":
            isSum = False
            isMult = False
            isDiv = False
        if op == "*":
            isSum = False
            isSub = False
            isDiv = False
        if op == "/" or op == "\\":
            isSum = False
            isSub = False
            isMult = False
    isOp = isSum or isSub or isMult or isDiv

    if isOp:
        results.write("the")
        results.write(" ")

        if isSum:
            results.write("sum")
            results.write(" ")
        elif isSub:
            results.write("difference")
            results.write(" ")
        elif isMult:
            results.write("product")
            results.write(" ")

        results.write("between")
        results.write(" ")

        results.write(generate_vars_symbols(body, symbols, arithAtom))

    return results.getvalue()


def generate_compare_of_arithmetic_builtin(op, compareTerm):
    results = StringIO()

    results.write(" ")
    results.write("is")
    results.write(" ")
    results.write(generate_compare_operator_sentence(op))
    results.write(" ")

    results.write(compareTerm.toString())

    return results.getvalue()


def generate_aggregate_subsentence(aggregate, symbols, costWeakTerm=None, isStrongConstraint=True):
    results = StringIO()
    # #AGGR{VL, X: scoreassignment(X,VL)} = 1
    # --->
    # the lowest value of a scoreAssignment with movie id X for each id is equal to 1

    operator = None
    assignmentVar = None
    operator = getAggregateOperator(aggregate)
    if operator == "=":
        if (aggregate.lowerOp == "=" and aggregate.upperGuard is None) or (
                aggregate.upperOp == "=" and aggregate.lowerGuard is None):
            if aggregate.upperGuard is not None and aggregate.upperGuard.isVariable():
                assignmentVar = aggregate.upperGuard
            elif aggregate.lowerGuard is not None and aggregate.lowerGuard.isVariable():
                assignmentVar = aggregate.lowerGuard

    if operator is not None:
        if (costWeakTerm is not None and assignmentVar is not None and costWeakTerm.name == assignmentVar.name
                or isStrongConstraint):
            results.write("")
        else:
            if costWeakTerm is not None:
                results.write("whenever")
                results.write(" ")
                results.write("we have that")
                results.write(" ")

    for aggregateElement in aggregate.aggregateElement:
        aggrTerm = aggregateElement.leftTerms[0]
        forEachTerms = aggregateElement.leftTerms[1:]
        forEachSubsentences = None

        if len(aggregateElement.leftTerms) > 1:
            forEachSubsentences = [None] * (len(aggregateElement.leftTerms) - 1)
        foundClassicalLiteral = None
        foundVarOfLiteral = None
        positionOfFoundVar = -1
        foundMultipleUseOfAggrTerm = False
        aggrVarsWithMultipleUse = []
        builtinAtoms = []
        removeAggrTerm = True

        # Search builtins
        for naf_literal in aggregateElement.body.literals:
            if type(naf_literal) == NafLiteral and type(naf_literal.literal) == BuiltinAtom:
                # if type(lit.literal.terms[0]) == ArithmeticAtom:
                #    hasSumInBuiltin = True
                builtinAtoms.append(naf_literal.literal)
                if naf_literal.literal.containsVar(aggrTerm):
                    removeAggrTerm = False

        for naf_literal in aggregateElement.body.literals:
            # if foundClassicalLiteral is None:
            if type(naf_literal) == NafLiteral:
                if type(naf_literal.literal) == ClassicalLiteral:
                    p = 0
                    for t in naf_literal.literal.terms:
                        if t.name == aggrTerm.name:
                            if foundClassicalLiteral:
                                foundMultipleUseOfAggrTerm = True
                            else:
                                foundClassicalLiteral = naf_literal.literal
                                foundVarOfLiteral = t
                                positionOfFoundVar = p
                            foundAMatchWithAggregateTerm = True
                        if t in forEachTerms:
                            subEach = StringIO()
                            subEach.write("for each")
                            subEach.write(" ")
                            symbLit = get_symbol(symbols, naf_literal.literal)
                            subEach.write(symbLit.attributes[p])
                            if forEachSubsentences[forEachTerms.index(t)] != None:
                                subEach.write(" ")
                                subEach.write(t.name)
                            forEachSubsentences[forEachTerms.index(t)] = subEach.getvalue()
                        p = p + 1

    connective = None
    if foundClassicalLiteral is not None:
        results.write("the")
        results.write(" ")
        if aggregate.aggregateFunction == "#min":
            results.write("lowest")
            connective = "of"
        elif aggregate.aggregateFunction == "#max":
            results.write("highest")
            connective = "of"
        elif aggregate.aggregateFunction == "#count":
            results.write("number of")
            connective = "that have"
        elif aggregate.aggregateFunction == "#sum":
            results.write("total")
            connective = "that have"
            # connective = "of"
        results.write(" ")
        symbLit = get_symbol(symbols, foundClassicalLiteral)
        results.write(symbLit.attributes[positionOfFoundVar])
        # results.write(" ")

        if foundMultipleUseOfAggrTerm:
            results.write(" ")
            results.write(aggrTerm.name)

        if forEachSubsentences is not None:
            for s in forEachSubsentences:
                if s is not None:
                    results.write(", ")
                    results.write(s)
                    results.write(",")
        # else:
        #    results.write(" ")

        if connective is not None:
            results.write(" ")
            results.write(connective)
        results.write(" ")
        results.write("a")
        results.write(" ")
        results.write(foundClassicalLiteral.name)
        results.write(" ")

        tmpLitTerm = None
        tmpSymbTerm = None
        if removeAggrTerm:
            tmpLitTerm = foundClassicalLiteral.terms.pop(positionOfFoundVar)
            tmpSymbTerm = symbLit.attributes.pop(positionOfFoundVar)
        results.write(generate_with(foundClassicalLiteral, symbLit, builtinAtoms))
        # results.write(" ")
        if removeAggrTerm:
            foundClassicalLiteral.terms.insert(positionOfFoundVar, tmpLitTerm)
            symbLit.attributes.insert(positionOfFoundVar, tmpSymbTerm)

        if len(aggregateElement.body.literals) > 1:
            # results.write("in")
            results.write(" ")
            startedIn = False
            for nafLit in aggregateElement.body.literals:
                if nafLit.literal != foundClassicalLiteral:
                    if type(nafLit) == NafLiteral and type(nafLit.literal) == ClassicalLiteral:
                        if startedIn:
                            results.write(",")
                            results.write(" ")
                        else:
                            startedIn = True
                        results.write(nafLit.literal.name)
                        results.write(" ")
                        results.write(nafLit.literal.terms[0].name)
                        symbLit2 = get_symbol(symbols, nafLit.literal)
                        results.write(" ")
                        results.write(generate_with(nafLit.literal, symbLit2, builtinAtoms))
            # results.write(" ")

        if operator is not None:
            if costWeakTerm is not None and assignmentVar is not None and costWeakTerm.name == assignmentVar.name:
                results.write("")
            else:
                results.write(" ")
                results.write("is")
                results.write(" ")
                results.write(generate_compare_operator_sentence(operator))
                results.write(" ")
                if operator == "between":
                    results.write(aggregate.lowerGuard.toString())
                    results.write(" ")
                    results.write("and")
                    results.write(" ")
                    results.write(aggregate.upperGuard.toString())
                else:
                    if aggregate.lowerGuard is not None:
                        results.write(aggregate.lowerGuard.toString())
                    else:
                        results.write(aggregate.upperGuard.toString())
    results.write(generateWhereForBuiltins(builtinAtoms))
    return results.getvalue()


'''
    @dataclass(frozen=True)
class AggregateElement:
    leftTerms: list[Term]
    body: 'Conjunction'

@dataclass(frozen=True)
class Aggregate:
    lowerGuard: Term  
    upperGuard: Term
    lowerOp: str
    upperOp: str
    aggregateFunction: str
    aggregateElement: list[AggregateElement]
'''



from textwrap import dedent

from setuptools import setup, find_packages

setup(
    name='asp2cnl',
    version='1.0.2',
    description='A tool for converting Answer Set Programming into a Controlled Natural Language based on English',
    long_description=dedent('''\
                            A tool for converting Answer Set Programming into a Controlled Natural Language based on English

                            '''),
    url='https://github.com/dodaro/asp2cnl',
    license='Apache License',
    author='Carmine Dodaro',
    author_email='carmine.dodaro@unical.it',
    maintainer='Kristian Reale',
    maintainer_email='kristian.reale@unical.it',
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['test*']),
    package_data={'': ['asp_core_2_grammar/asp_grammar.lark']},
    entry_points={
        'console_scripts': ['asp2cnl = asp2cnl.asp2cnl:run_asp2cnl',
                            'asp2nl = asp2nl.asp2nl:run_asp2nl',
                            'cnl2nl = cnl2nl.cnl2nl:run_cnl2nl'],
    },
    install_requires=['lark', 'inflect', 'multipledispatch', 'cnl2asp', 'requests'],
    python_requires=">=3.10"
)


import argparse
import os

import requests
import json

OLLAMA_URL = "http://localhost:11434/api/generate"


def compile_prompt(prefix="Explain in everyday language what the following sentence says, improving the writing "
                          "style:\n", suffix="\nLimit your answer to 500 characters."):
    """
    Helper method to compile a prompt with both fixed prefix and suffix
    :param prefix:
    :param suffix:
    :return:
    """

    def _inner(cnl):
        prompt_template = prefix + cnl + suffix
        return prompt_template

    return _inner


def make_request_to_service(service_url=OLLAMA_URL, data=None):
    response = requests.post(service_url, data=data, headers={"Content-Type": "application/json"})
    return json.loads(response.text)


def contact_ollama_service(model='openchat', prompt=None):
    json_response = make_request_to_service(data=json.dumps({"model": model, "prompt": prompt, "stream": False}))
    return json_response['response']


def contact_openai_service(prompt, model="gpt-3.5-turbo"):
    try:
        from openai import OpenAI
        client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"), )
    except:
        return "NO OPENAI_API_KEY"
    messages = [{"role": "user", "content": prompt}]
    response = client.chat.completions.create(model=model, messages=messages, temperature=0, )
    return response.choices[0].message["content"]


def contact_llm_service(model='openchat', prompt=None):
    if model == 'openai':
        return contact_openai_service(prompt)
    else:
        return contact_ollama_service(model=model, prompt=prompt)


def run_cnl2nl():
    parser = argparse.ArgumentParser()
    parser.add_argument('--cnl2nl', action='store_true')
    parser.add_argument("-m", "--llm_model", type=str, default='openchat')
    parser.add_argument('cnl_file')
    parser.add_argument('output_file', type=str, nargs='?', default='')
    args = parser.parse_args()

    cnl_file = args.cnl_file
    llm_model = args.llm_model
    output_file = args.output_file


    prompt_compiler = compile_prompt()

    with open(cnl_file, "r") as f:
        o = json.load(f)
    o['nl'] = []
    for cnl in o['cnl']:
        print(f'Processing rule: {cnl}')
        o['nl'].append(contact_llm_service(model=llm_model, prompt=prompt_compiler(cnl)))

    print("\n** NL:")
    print('\n'.join(o['nl']))

    try:
        with open(output_file, 'w') as f:
            json.dump(o, f)
    except Exception as e:
        print("Error in writing output", str(e))



from lark.lexer import Token

from lark import Transformer, Lark
from dataclasses import dataclass

from io import StringIO
import os


class ASPParser:
    __aspCoreParser = Lark(
        open(os.path.join(os.path.dirname(__file__), "asp_core_2_grammar/asp_grammar.lark"), "r").read())

    def __init__(self, programFile: str):
        self.__programFile = programFile

    def parse(self):
        parsed = self.__aspCoreParser.parse(self.__programFile)
        #print(parsed)
        content_tree: ASPContentTree = ASPTransformer().transform(parsed)
        definitions = [content_tree.rules[i] for i in range(len(content_tree.rules))]
        #print(content_tree)
        return definitions


class ASPTransformer(Transformer):
    def __init__(self):
        super().__init__()
        self.__rules_list = []

    def start(self, elem):
        tree = ASPContentTree(self.__rules_list[:])
        self.__rules_list = []
        return tree

    def head(self, elem):
        if len(elem) == 1 and type(elem[0]) == Choice:
            return elem[0]
        else:
            disElements = []
            for e in elem[0]:
                if type(e) == ClassicalLiteral:
                    disElements.append(e)
            return Disjunction(disElements)

    def disjunction(self, elem):
        disElements = []
        for e in elem:
            if type(e) == ClassicalLiteral:
                disElements.append(e)
            else:
                for e1 in e:
                    if type(e1) == ClassicalLiteral:
                        disElements.append(e1)
        return disElements

    def classical_literal(self, elem):
        classicalLit = None
        if len(elem) == 1:
            classicalLit = ClassicalLiteral(elem[0].value, [])
        else:
            classicalLit = ClassicalLiteral(elem[0].value, elem[2][:])
        return classicalLit

    def aggregate(self, elem):
        lowerGuard = None
        upperGuard = None
        lowerOp = None
        upperOp = None
        aggregateFunction = None
        foundCurlyOpen = False
        foundCurlyClose = False
        aggregateElements = None
        for e in elem:
            if type(e) == list:
                if not foundCurlyOpen:
                    lowerGuard = e[0]
                    lowerOp = e[1]
                else:
                    aggregateElements = e
            elif self.__isAggrageteFunction__(e):
                aggregateFunction = e
            elif e == '_CURLY_OPEN_':
                foundCurlyOpen = True
            elif e == '_CURLY_CLOSE_':
                foundCurlyClose = True
            elif type(e) == str and foundCurlyClose:
                upperOp = e
            elif (type(e) == Term or type(e) == ArithmeticAtom) and foundCurlyClose:
                upperGuard = e

        return AggregateLiteral(lowerGuard, upperGuard, lowerOp, upperOp, aggregateFunction, aggregateElements)

    def aggregate_function(self, elem):
        return elem[0].value

    def aggregate_elements(self, elem):
        aggElements = []
        for e in elem:
            if type(e) == AggregateElement:
                aggElements.append(e)
            else:
                for e1 in e:
                    if type(e1) == AggregateElement:
                        aggElements.append(e1)
        return aggElements

    def aggregate_element(self, elem):
        aggregateTerms = elem[0]
        aggregateBody = Conjunction(elem[2])
        return AggregateElement(aggregateTerms, aggregateBody)

    def naf_literals(self, elem):
        lits = []
        for t in elem:
            if type(t) == NafLiteral:
                lits.append(t)
            else:
                for t1 in t:
                    if type(t1) == NafLiteral:
                        lits.append(t1)
        return lits

    def NAF(self, elem):
        return "_NOT_"

    def HASHTAG(self, elem):
        return "_HASHTAG_"

    def naf_literal(self, elem):
        thereIsNaf = False
        lit = None
        for e in elem:
            if e == "_NOT_":
                thereIsNaf = True
            elif type(e) == ClassicalLiteral or type(e) == BuiltinAtom:
                lit = e
        nafLit = NafLiteral(thereIsNaf, lit)
        return nafLit

    def binop(self, elem):
        return elem[0]

    def EQUAL(self, elem):
        return elem.value

    def GREATER(self, elem):
        return elem.value

    def GREATER_OR_EQ(self, elem):
        return elem.value

    def LESS(self, elem):
        return elem.value

    def LESS_OR_EQ(self, elem):
        return elem.value

    def UNEQUAL(self, elem):
        return elem.value

    def builtin_atom(self, elem):
        return BuiltinAtom(elem[0][1], [elem[0][0], elem[1]])

    def b(self, elem):
        return elem

    def terms(self, elem):
        terms = []
        for t in elem:
            if type(t) == Term or type(t) == ArithmeticAtom:
                terms.append(t)
            else:
                for t1 in t:
                    if type(t1) == Term or type(t1) == ArithmeticAtom:
                        terms.append(t1)
        return terms

    def term(self, elem):
        if len(elem) == 1:
            return Term(elem[0].value)
        elif len(elem) == 2:
            #if type(elem[1]) == ArithmeticAtom:
            #    operators = [elem[0][1]]
            #    operators += elem[1].ops
            #    terms = [Term(elem[0][0].value)]
            #    terms += elem[1].terms
            #    return ArithmeticAtom(operators, terms)
            #else:
            if elem[0][0].value == "(" or elem[0][0].value == "|":
                isParenthesis = elem[0][0].value == "("
                isModule = elem[0][0].value == "|"
                #return elem[0][1]
                if isinstance(elem[0][1], ArithmeticAtom):
                    isModule = isModule or elem[0][1].isModule
                    elem[0][1] = ArithmeticAtom(elem[0][1].ops, elem[0][1].terms, isParenthesis, isModule=isModule)
                if type(elem[1]) == ArithmeticAtom and not elem[1]._ArithmeticAtom__inParenthesis and not elem[1].isModule:
                    operators = [elem[0][3]]
                    operators += elem[1].ops
                    terms = [elem[0][1]]
                    terms += elem[1].terms
                    return ArithmeticAtom(operators, terms)
                else:
                    return ArithmeticAtom([elem[0][3]], [elem[0][1], elem[1]])
            elif type(elem[1]) == ArithmeticAtom and not elem[1]._ArithmeticAtom__inParenthesis and not elem[1].isModule:
                operators = [elem[0][1]]
                operators += elem[1].ops
                terms = [Term(elem[0][0].value)]
                terms += elem[1].terms

                return ArithmeticAtom(operators, terms)
            else:
                return ArithmeticAtom(elem[0][1], [Term(elem[0][0].value), elem[1]])
        #elif len(elem) == 3:# and elem[0].value == "(" and elem[-1].value == ")":
        #operators = elem[1].ops
                #       print(operators)
                #terms = elem[1].terms
                #return ArithmeticAtom(operators, terms)
        elif len(elem) == 3:
            if type(elem[1]) == ArithmeticAtom:
                if elem[0].value == "(":
                    return ArithmeticAtom(elem[1].ops, elem[1].terms, True)
                if elem[0].value == "|" and elem[2].value == "|":
                    return ArithmeticAtom(elem[1].ops, elem[1].terms, True, True)
            else:
                if isinstance(elem[0], Token) and elem[0].value == "|" and elem[2].value == "|":
                    return Term(elem[1].toString(), isModule=True)
                elif elem[1].value == "..":
                    return Term(elem[0].toString(), elem[2].toString())

    def termdue(self, elem):
        return elem

    def COLON(self, elem):
        return "_COLON_"

    def CONS(self, elem):
        return "_IF_"

    def WCONS(self, elem):
        return "_WEAK_IF_"

    def AT(self, elem):
        return "_AT_"

    def MINUS(self, elem):
        return "_MINUS_"

    def CURLY_OPEN(self, elem):
        return "_CURLY_OPEN_"

    def CURLY_CLOSE(self, elem):
        return "_CURLY_CLOSE_"

    def choice(self, elem):
        lowerGuard = None
        upperGuard = None
        lowerOp = None
        upperOp = None
        choice_elements = None
        foundCurlyOpen = False
        foundCurlyClose = False
        for e in elem:
            if e == "_CURLY_OPEN_":
                foundCurlyOpen = True
            elif e == "_CURLY_CLOSE_":
                foundCurlyClose = True
            else:
                if type(e) == Term:
                    if not foundCurlyOpen:
                        lowerGuard = e
                    elif foundCurlyClose:
                        upperGuard = e
                elif type(e) == str:
                    if not foundCurlyOpen:
                        lowerOp = e
                    elif foundCurlyClose:
                        upperOp = e
                else:
                    choice_elements = e
        return Choice(lowerGuard, upperGuard, lowerOp, upperOp, choice_elements)

    def choice_elements(self, elem):
        choiceElements = []
        for e in elem:
            if type(e) == ChoiceElement:
                choiceElements.append(e)
            else:
                for e1 in e:
                    if type(e1) == ChoiceElement:
                        choiceElements.append(e1)
        return choiceElements

    def choice_element(self, elem):
        left_part = None
        right_part = None
        foundColon = False
        for e in elem:
            if type(e) == ClassicalLiteral:
                left_part = e
            elif e == "_COLON_":
                foundColon = True
            # elif type(e) == NafLiteral:
            elif type(e) == list:
                if foundColon:
                    right_part = e

        return ChoiceElement(left_part, right_part)

    def arithop(self, elem):
        if (elem[0] == "_MINUS_"):
            return "-"
        return elem[0].value

    def weight_at_level(self, elem):
        beforeAtTerm = None
        afterAtTerms = None
        foundAt = False
        foundMinus = False
        for e in elem:
            if e == "_AT_":
                foundAt = True
            elif e == "_MINUS_":
                foundMinus = True
            else:
                if not foundAt:
                    beforeAtTerm = e
                else:
                    eToCheck = []
                    if type(e) == list:
                        eToCheck = e
                    else:
                        eToCheck.append(e)
                    for e1 in eToCheck:
                        if type(e1) == Term:
                            if afterAtTerms is None:
                                afterAtTerms = []
                            afterAtTerms.append(e1)
        return WeakElement(beforeAtTerm, afterAtTerms, foundMinus)

    def body(self, elem):
        bodyElements = []

        for e in elem:
            if type(e) == NafLiteral or type(e) == AggregateLiteral:
                bodyElements.append(e)
            else:
                for e1 in e:
                    if type(e1) == NafLiteral or type(e1) == AggregateLiteral:
                        bodyElements.append(e1)

        return bodyElements

    def body_choice_suchthat(self, elem):
        bodyElements = []

        for e in elem:
            if type(e) == NafLiteral:
                bodyElements.append(e)
            elif type(e) == list:
                for e1 in e:
                    if type(e1) == NafLiteral:
                        bodyElements.append(e1)

        return bodyElements

    def directive(self, elem):
        return Directive(elem[1].value, elem[2].value, elem[4].value)

    def statement(self, elem):
        if type(elem[0]) == Directive:
            self.__rules_list.append(elem[0])
        else:
            foundIf = False
            head = None
            body = None
            weakElement = None
            for e in elem:
                if type(e) == Disjunction or type(e) == Choice:
                    head = e
                elif e == "_IF_" or e == "_WEAK_IF_":
                    foundIf = True
                elif type(e) == list:
                    if foundIf:
                        body = Conjunction(e)
                elif type(e) == WeakElement:
                    weakElement = e

            self.__rules_list.append(Rule(head, body, weakElement))
        return elem

    def __isAggrageteFunction__(self, value):
        return value == "#min" or value == "#max" or value == "#sum" or value == "#count"  # ecc


@dataclass(frozen=True)
class Term:
    name: str
    afterDotDot: str = None
    isModule: bool = False


    def isVariable(self):
        return not self.name.isnumeric() and self.name[0].isupper() and self.afterDotDot is None

    def isUnderscore(self):
        return self.name == "_" and self.afterDotDot is None

    def isWithDotDot(self):
        return self.afterDotDot is not None

    def toString(self):
        if self.afterDotDot is None:
            text = StringIO()
            if self.isModule:
                text.write("|")
            text.write(self.name)
            if self.isModule:
                text.write("|")
            return text.getvalue()
        else:
            return self.name + ".." + self.afterDotDot


@dataclass(frozen=True)
class ArithmeticAtom:
    ops: list[str]
    terms: list[Term]
    __inParenthesis:bool = False
    isModule: bool = False

    def containsVar(self, term):
        for t in self.terms:
            if t.name == term.name:
                return True
        return False

    def isVariable(self):
        return False

    def toString(self):
        text = StringIO()

        if self.isModule:
            text.write("|")
        for i in range(len(self.ops)):
            if type(self.terms[i]) == ArithmeticAtom and not self.terms[i].isModule:
                text.write("(")
            text.write(self.terms[i].toString())
            if type(self.terms[i]) == ArithmeticAtom and not self.terms[i].isModule:
                text.write(")")
            text.write(self.ops[i])
        if type(self.terms[len(self.terms) - 1]) == ArithmeticAtom and not self.terms[len(self.terms) - 1].isModule:
            text.write("(")
        text.write(self.terms[len(self.terms) - 1].toString())
        if type(self.terms[len(self.terms) - 1]) == ArithmeticAtom and not self.terms[len(self.terms) - 1].isModule:
            text.write(")")
        if self.isModule:
            text.write("|")
        return text.getvalue()
        # return self.terms[0].toString() + self.ops[0] + self.terms[1].toString()


@dataclass(frozen=True)
class BuiltinAtom:
    op: str
    terms: list[Term | ArithmeticAtom]

    def containsVar(self, term):
        for t in self.terms:
            if type(t) == ArithmeticAtom:
                if t.containsVar(term.name):
                    return True
            else:
                if t.name == term.name:
                    return True
        return False

    def toString(self):
        #if type(self.terms[0]) == Token and type(self.terms[-1]) == Token and self.terms[0].value == "|" and self.terms[-1].value == "|":
        #    return self.terms[1].toString() + " " + self.op + " " + self.terms[2].toString()
        text = StringIO()
        #if self.isModule:
        #    text.write("|")
        text.write(self.terms[0].toString() + " " + self.op + " " + self.terms[1].toString())
        #if self.isModule:
        #    text.write("|")
        return text.getvalue()


@dataclass(frozen=True)
class ClassicalLiteral:
    name: str
    terms: list[Term]

    def arity(self):
        return len(self.terms)

    def hasVariables(self):
        for term in self.terms:
            if term.isVariable():
                return True

    def __hash__(self):
        return hash(repr(self))
    def toString(self):
        text = StringIO()
        text.write(self.name)
        started = False
        for t in self.terms:
            if started:
                text.write(", ")
            else:
                text.write("(")
                started = True
            text.write(t.toString())
        if started:
            text.write(")")
        return text.getvalue()


@dataclass(frozen=True)
class NafLiteral:
    isNot: bool
    literal: ClassicalLiteral | BuiltinAtom

    def __hash__(self):
        return hash(repr(self))

    def toString(self):
        text = ""
        if self.isNot:
            text = "not "
        text = text + self.literal.toString()
        return text




@dataclass(frozen=True)
class AggregateElement:
    leftTerms: list[Term]
    body: 'Conjunction'

    def toString(self):
        text = StringIO()
        started = False
        for t in self.leftTerms:
            if started:
                text.write(",")
                text.write(" ")
            else:
                started = True
            text.write(t.toString())
        text.write(":")
        text.write(" ")
        text.write(self.body.toString())
        return text.getvalue()


@dataclass(frozen=True)
class AggregateLiteral:
    lowerGuard: Term
    upperGuard: Term
    lowerOp: str
    upperOp: str
    aggregateFunction: str
    aggregateElement: list[AggregateElement]

    def toString(self):
        text = StringIO()
        if self.lowerGuard is not None:
            text.write(self.lowerGuard.toString())
            text.write(" ")
            text.write(self.lowerOp)
            text.write(" ")
        text.write(self.aggregateFunction)
        text.write("{")
        started = False
        for aggrEl in self.aggregateElement:
            if started:
                text.write(";")
                text.write(" ")
            else:
                started = True
            text.write(aggrEl.toString())
        text.write("}")
        if self.upperGuard is not None:
            text.write(" ")
            text.write(self.upperOp)
            text.write(" ")
            text.write(self.upperGuard.toString())
            text.write(" ")
        return text.getvalue()


@dataclass(frozen=True)
class Conjunction:
    literals: list[NafLiteral | AggregateLiteral]

    def hasVariables(self):
        for lit in self.literals:
            for term in lit.classical_literal.terms:
                if term.isVariable():
                    return True
        return False

    def toString(self):
        text = StringIO()
        started = False
        for l in self.literals:
            if started:
                text.write(", ")
            else:
                started = True
            text.write(l.toString())
        return text.getvalue()


@dataclass(frozen=True)
class ChoiceElement:
    left_part: ClassicalLiteral
    right_part: list[NafLiteral]

    def toString(self):
        text = StringIO()
        text.write(self.left_part.toString())
        if self.right_part is not None:
            text.write(":")
            started = False
            for nafLit in self.right_part:
                if started:
                    text.write(",")
                    text.write(" ")
                else:
                    started = True
                text.write(nafLit.toString())
        return text.getvalue()


@dataclass(frozen=True)
class Choice:
    lowerGuard: Term
    upperGuard: Term
    lowerOp: str
    upperOp: str
    elements: list[ChoiceElement]

    def toString(self):
        text = StringIO()
        if self.lowerGuard is not None:
            text.write(self.lowerGuard.name)
            text.write(" ")
            text.write(self.lowerOp)
            text.write(" ")
        text.write("{")
        startedElems = False
        for ce in self.elements:
            if startedElems:
                text.write(";")
            else:
                startedElems = True
            text.write(ce.toString())
        text.write("}")
        if self.upperGuard is not None:
            text.write(" ")
            text.write(self.upperOp)
            text.write(" ")
            text.write(self.upperGuard.name)
        return text.getvalue()


@dataclass(frozen=True)
class Disjunction:
    atoms: list[ClassicalLiteral]

    def hasVariables(self):
        for atom in self.atoms:
            for term in atom.terms:
                if term.isVariable():
                    return True
        return False

    def toString(self):
        text = StringIO()
        started = False
        for a in self.atoms:
            if started:
                text.write(" | ")
            else:
                started = True
            text.write(a.toString())
        return text.getvalue()


@dataclass(frozen=True)
class WeakElement:
    beforeAt: Term
    afterAt: list[Term]
    isMaximize: bool = False

    def toString(self):
        text = StringIO()
        text.write("[")
        if self.beforeAt is not None:
            if (self.isMaximize):
                text.write("-")
            text.write(self.beforeAt.toString())
        if self.afterAt is not None:
            text.write("@")
            started = False
            for elem in self.afterAt:
                if started:
                    text.write(",")
                else:
                    started = True
                text.write(elem.toString())
        text.write("]")
        return text.getvalue()


@dataclass(frozen=True)
class Directive:
    type: str
    name: str
    value: int

    def toString(self):
        text = StringIO()
        text.write("#")
        text.write(self.type)
        text.write(" ")
        text.write(self.name)
        text.write(" = ")
        text.write(self.value)
        return text.getvalue()


@dataclass(frozen=True)
class Rule:
    head: Disjunction | Choice
    body: Conjunction
    weight_at_level: WeakElement = None

    def isFact(self):
        return type(self.head) != Choice and self.head is not None and len(
            self.head.atoms) == 1 and self.body is None and not self.head.atoms[0].hasVariables()

    def isClassical(self):
        return type(self.head) != Choice and self.head is not None and len(self.head.atoms) == 1 and not \
        self.head.atoms[0] == Choice and self.body is not None and len(self.body.literals) > 0

    def isStrongConstraint(self):
        return type(self.head) != Choice and self.head is None and self.body is not None and len(
            self.body.literals) > 0 and self.weight_at_level is None

    def isDisjunctive(self):
        return type(self.head) != Choice and self.head is not None and len(
            self.head.atoms) > 1 and self.body is not None and len(self.body.literals) > 0

    def isChoice(self):
        return type(self.head) == Choice and self.body is not None and len(self.body.literals) > 0

    def isWeakConstraint(self):
        return type(self.head) != Choice and self.head is None and self.body is not None and len(
            self.body.literals) > 0 and self.weight_at_level is not None

    def toString(self):
        text = StringIO()
        if self.head is not None:
            text.write(self.head.toString())
        if self.body is None:
            text.write(".")
        else:
            if self.isWeakConstraint():
                text.write(" :~ ")
            else:
                text.write(" :- ")
            text.write(self.body.toString())
        text.write(".")
        if self.isWeakConstraint():
            text.write(" ")
            text.write(self.weight_at_level.toString())
        return text.getvalue()


@dataclass(frozen=True)
class ASPContentTree:
    rules: list
